unit Decompressionthread;

interface

uses
  dialogs,Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  ScktComp2,zlib,cyclicBuffer,generic;

//const
   //bufsize = 1024 * 1024 * 5;

type
  TDecompressionThread = class(TThread)
  private
   audioOutputBuffer,compressedBuffer: TCyclicBuffer;
   frmOwner: TForm;
  protected
    procedure Execute; override;
  public
    constructor create(_frmOwner: TForm);
    destructor Destroy; override;
    function getBytesForSpeaker(buffer: pchar; var size: integer): boolean;
    procedure HandleNewBytes(buffer: pbyte; size: integer);
  end;

implementation

uses unit1;



{constructor}
constructor TDecompressionThread.create(_frmOwner: TForm);
begin
   inherited create(true); // create but don't start running yet
   frmOwner := _frmOwner;

   audioOutputBuffer := TCyclicBuffer.create(bufsize);
   compressedBuffer := TCyclicBuffer.create(bufsize);
end;

{destructor}
destructor TDecompressionThread.Destroy;
begin
   audioOutputBuffer.free;
   compressedBuffer.free;
   inherited destroy;
end;


{main thread entry point}
procedure TDecompressionThread.Execute;
var
   uncompressedBuffer: array[0..audioBufferSize-1] of byte;
   tmpCompressedBuf: array of byte;
   numberOfCompressedBytes,numberOfUncompressedBytes: integer;
   compressedBytesStream: TMemoryStream;
   decompressionStream: TDecompressionStream;
   checksum, receivedCompressedChecksum, receivedUncompressedChecksum: byte;
begin
   while (not terminated) and (connected) do
   begin
      try
         // uncompress bytes from compressedBuffer and place them into the audio output buffer
         if compressedBuffer.available < 2 then
         begin
            logToFile('Only ' + inttostr(compressedBuffer.available) + ' compressed bytes available');
            while compressedBuffer.available < 2 do
            begin
               sleep(100);
               if (terminated) or (not connected) then exit;
            end;
         end;

         // read the number of compressed bytes
         // this is a 16-bit number consisting of the first two bytes
         logToFile('Reading number of compressed bytes');
         numberOfCompressedBytes := compressedBuffer.popIntegerLittleEndian(2);
         logToFile('Found block of ' + inttostr(numberOfCompressedBytes) + ' compressed bytes');

         // the next 2 byte are the checksums of the compressed bytes
         // and the uncompressed bytes
         receivedCompressedChecksum := compressedBuffer.pop;
         receivedUncompressedChecksum := compressedBuffer.pop;

         // check we have enough available bytes in compressedBuffer
         if compressedBuffer.available < numberOfCompressedBytes then
         begin
            logToFile('Only ' + inttostr(compressedBuffer.available) + ' bytes in compressedBuffer');
            while compressedBuffer.available < numberOfCompressedBytes do
            begin
               sleep(100);
               if (terminated) or (not connected) then exit;
            end;
         end;
            
         // we need a temporary buffer to decompress the bytes
         setLength(tmpCompressedBuf,numberOfCompressedBytes);
         //assert(numberOfCompressedBytes <= sizeof(tmpCompressedBuf));
         compressedBuffer.popBuffer(@tmpCompressedBuf[0], numberOfCompressedBytes);
         logBytes('compressed bytes:', @tmpCompressedBuf[0], numberOfCompressedBytes);

         // make sure the compressed checksum matches
         checksum := calcChecksum(@tmpCompressedBuf[0], numberOfCompressedBytes);
         logToFile('Calculated compressed checksum:' + inttostr(checksum) + ', received checksum:' + inttostr(receivedCompressedChecksum));
         assert(checksum = receivedCompressedChecksum);

         // decompress the bytes
         compressedBytesStream := TMemoryStream.Create;
         compressedBytesStream.write(tmpCompressedBuf[0],numberOfCompressedBytes);
         compressedBytesStream.Seek(soFromBeginning,0);
         decompressionStream := TDecompressionStream.Create(compressedBytesStream);
         numberOfUncompressedBytes := decompressionStream.read(uncompressedBuffer, sizeof(uncompressedBuffer));
         assert(numberOfUncompressedBytes = sizeof(uncompressedBuffer));
         decompressionStream.Free;
         compressedBytesStream.free;
         tmpCompressedBuf := nil;
         logBytes('uncompressed bytes:', @uncompressedBuffer, numberOfUncompressedBytes);

         // make sure the uncompressed checksum matches
         checksum := calcChecksum(@uncompressedBuffer[0], numberOfUncompressedBytes);
         logToFile('Calculated uncompressed checksum:' + inttostr(checksum) + ', received checksum:' + inttostr(receivedUncompressedChecksum));
         assert(checksum = receivedUncompressedChecksum);

         // now we have the original uncompressed bytes, write them into the audio output buffer
         logToFile('Writing ' + inttostr(numberOfUncompressedBytes) + ' into audio output buffer');
         audioOutputBuffer.pushBuffer(@uncompressedBuffer[0], numberOfUncompressedBytes);
      except
         on E: Exception do
         begin
            logToFile('Error in decompression thread: ' + E.message);
         end;
      end;
   end;
end ;


{called by TAudioOut on the main form when it needs some bytes to play out through the speaker}
function TDecompressionThread.getBytesForSpeaker(buffer: pchar; var size: integer): boolean;
begin
   Result := connected; // tell AudioOut to read more buffers

   if audioOutputBuffer.available < audioBufferSize then   // fixme
   begin
      logToFile('Speaker wants bytes, but there are only ' +inttostr(audioOutputBuffer.available) + ' available');
      sleep(300);
      size := 0;
      exit;
   end;

   size := audioBufferSize;

   // fill the audio buffer with bytes from the transfer buffer
   // if bit16Sound and stereo then assert(size mod 4 = 0); // buffer size must be divisible by 4 - see TAudio help file
   logToFile('Speaker fetching ' + inttostr(size) + '/' + inttostr(audioOutputBuffer.available) + ' bytes from audio buffer');

   // copy bytes from the audio buffer into the pchar buffer used by TAudioOut
   audioOutputBuffer.popBuffer(pbyte(buffer), size);
   logBytes('Speaker playing:', pbyte(buffer), size);
end;


{called when a block of decrypted, compressed bytes arrives from the receiver thread}
procedure TDecompressionThread.HandleNewBytes(buffer: pbyte; size: integer);
begin
   logBytes('Pushing bytes to compressedBuffer:', buffer, size);
   compressedBuffer.pushBuffer(buffer, size);
end;


end.

